
#include <x4g/parser.h>
#include <x4g/constraints.h>

#include <gecode/search.hh>


class x4gSpace : public gec::Space
 {		 
   protected:	 
      gec::IntVarArray  gVars;

   public:   
      x4gSpace ()
         : gec::Space()
          { }

      x4gSpace (bool  share, x4gSpace  & xs)
         : gec::Space(share, xs)
          { gVars.update(*this, share, xs.gVars); }

      virtual gec::Space *  copy (bool  share)
       { return new x4gSpace(share, *this); }

      void setBranch ()
       {
         gec::branch(*this, this->gVars, gec::INT_VAR_SIZE_MIN,
                     gec::INT_VAL_MIN);
       }

      void print () const
       {
         std::cout << "v";

         for ( int i = 0; i < this->gVars.size() ; ++i )
            std::cout << " " << this->gVars[i];

         std::cout << std::endl;

         std::cout.flush();
       }

      gec::Space *  getSpacePnt () 
       { return(this); }

      gec::IntVarArray *  getGVars ()
       { return(&(this->gVars)); }
 };


int main (int      argc,
          char **  argv)
 {
   int ret(EXIT_FAILURE);

   static std::string const  preUn("unsupported ");

   try
    {
      if (argc!=2)
       {
         std::string  errStr("syntax: "); 

         errStr += argv[0];
         errStr += " CSPFile.xml";

         throw std::runtime_error(errStr);
       }

      x4gSpace *       pSp(new x4gSpace());
      x4g::XCSPParser  par(pSp->getSpacePnt(), pSp->getGVars());

      cxp::XMLParser_libxml2<>  parser(par);

      parser.parse(argv[1]); // parse the input file

      pSp->setBranch();

      gec::Search::Options  gso;

      gso.threads = 0.0;  // # of threads == # of CPUs

      gec::DFS<x4gSpace>  e(pSp, gso);

      delete pSp;

      x4gSpace *  s(e.next());

      if ( NULL != s )
       {
         std::cout << "s SATISFIABLE" << std::endl;

         s->print();

         delete s;
       }
      else
         std::cout << "s UNSATISFIABLE" << std::endl;

      ret = EXIT_SUCCESS;
    }
   catch (std::exception &e)
    {
      std::string const  wh(e.what());

      if ( 0 == wh.compare(0, preUn.length(), preUn) )
         std::cout << "s UNSUPPORTED" << std::endl;
      else
         std::cout << "s UNKNOWN" << std::endl;

      std::cout.flush();

      std::cerr << "\n\tException :\n";
      std::cerr << "\t" << wh << std::endl << std::endl;

      std::cerr.flush();
    }
   catch (...)
    {
      std::cout << "s UNKNOWN" << std::endl;

      std::cout.flush();

      std::cerr << "\n\tNon Standard Exception" << std::endl;

      std::cerr.flush();
    }

   return(ret);
 }

