
#include <set>

#include <x4g/constraints.h>

#include <gecode/scheduling.hh>


namespace x4g
 {
   std::map<NodeType, gec::BoolOpType>  constr::initMOpBT ()
    {
      std::map<NodeType, gec::BoolOpType>  ret;

      ret[F_AND] = gec::BOT_AND;
      ret[F_OR]  = gec::BOT_OR;
      ret[F_XOR] = gec::BOT_XOR;
      ret[F_IFF] = gec::BOT_EQV;

      return(ret);
    }

   std::map<NodeType, gec::IntRelType>  constr::initMOpRT ()
    {
      std::map<NodeType, gec::IntRelType>  ret;

      ret[F_EQ] = ret[SYMB_EQ] = gec::IRT_EQ;
      ret[F_NE] = ret[SYMB_NE] = gec::IRT_NQ;
      ret[F_GE] = ret[SYMB_GE] = gec::IRT_GQ;
      ret[F_GT] = ret[SYMB_GT] = gec::IRT_GR;
      ret[F_LE] = ret[SYMB_LE] = gec::IRT_LQ;
      ret[F_LT] = ret[SYMB_LT] = gec::IRT_LE;

      return(ret);
    }

   std::map<NodeType, gec::IntRelType>  constr::initMOpRTRev ()
    {
      std::map<NodeType, gec::IntRelType>  ret;

      ret[F_EQ] = ret[SYMB_EQ] = gec::IRT_EQ;
      ret[F_NE] = ret[SYMB_NE] = gec::IRT_NQ;
      ret[F_GE] = ret[SYMB_GE] = gec::IRT_LQ;
      ret[F_GT] = ret[SYMB_GT] = gec::IRT_LE;
      ret[F_LE] = ret[SYMB_LE] = gec::IRT_GQ;
      ret[F_LT] = ret[SYMB_LT] = gec::IRT_GR;

      return(ret);
    }


   std::map<NodeType, gec::BoolOpType> const
      constr::mOpBT(pred::initMOpBT());
   std::map<NodeType, gec::IntRelType> const
      constr::mOpRT(pred::initMOpRT());
   std::map<NodeType, gec::IntRelType> const
      constr::mOpRTRev(pred::initMOpRTRev());


   gec::BoolVar rel::getPostAnd (cxp::ASTList const      & args,
                                 gec::IntVarArray const  & gVars,
                                 gec::Space              & gSpace,
                                 int const *               tuple)
    {
      gec::BoolVar  ret;

      ret = gec::expr(
         gSpace,
         rel::getIntVarFromVar(gVars, args[1]) == tuple[1]);

      for ( int i = 2 ; i < args.nbArg() ; ++i )
         ret = gec::expr(
            gSpace,
            ret && (rel::getIntVarFromVar(gVars, args[i]) == tuple[i]));

      return(ret);
    }

   void rel::manageConf (cxp::ASTList const      & args,
                         gec::IntVarArray const  & gVars,
                         gec::Space              & gSpace) const
    {
      if ( 1 == this->arity )
       {
         for ( int i = 0 ; i < this->numTuples ; ++i )
            gec::rel(
               gSpace,
               rel::getIntVarFromVar(gVars, args[0]),
               gec::IRT_NQ,
               this->tSet[i][0]);
       }
      else
       {
         for ( int i = 0 ; i < this->numTuples ; ++i )
            gec::rel(
               gSpace,
               gec::expr(
                  gSpace,
                  rel::getIntVarFromVar(gVars, args[0])
                  == this->tSet[i][0]),
               gec::BOT_AND,
               this->getPostAnd(args, gVars, gSpace, this->tSet[i]),
               0);
       }
    }


   void rel::add (int    a0,
                  int *  t0,
                  int    c0)
    {
      if ( this->finalized )
         throw std::runtime_error("adding tuple in a finalized set");

      if ( this->arity != a0 )
         throw std::runtime_error("tuple arity error");

      this->tSet.add(gec::IntArgs(this->arity, t0));

      if ( REL_SOFT == this->relType )
         this->cV.push_back(c0);
    }

   void rel::finalize ()
    {
      if ( this->finalized )
         throw std::runtime_error("finalizing a finalized set");

      this->tSet.finalize();

      this->finalized = true;

      if ( this->numTuples != this->tSet.tuples() )
         throw std::runtime_error("tuple set cardinality error");
    }

   void rel::manage (cxp::ASTList const      & args,
                     gec::IntVarArray const  & gVars,
                     gec::Space              & gSpace) const
    {
      if ( false == this->finalized )
         throw std::runtime_error("relation finalize error");

      if ( args.nbArg() != this->arity )
         throw std::runtime_error("relation arity error");

      if ( REL_CONFLICT == this->relType )
         this->manageConf(args, gVars, gSpace);
      else if ( REL_SUPPORT == this->relType )
         this->manageSupp(args, gVars, gSpace);
      else if ( REL_SOFT == this->relType )
         this->manageSoft(args, gVars, gSpace);
      else
         throw std::runtime_error("unknown relation type");
    }

   void pred::checkAndSetAstFunc ()
    {
      if ( this->astTree->isOperator() )
       {
         this->astFunc
            = dynamic_cast<cxp::ASTAbstractFunction *>(this->astTree);

         if ( NULL == this->astFunc )
            throw std::runtime_error("dynamic cast failure");

         this->oper     = this->astFunc->getType();
         this->operArgs = this->getNArgs(this->oper);

         if ( false == this->isBooleanOp(this->oper) )
            throw std::runtime_error("non boolean operator");

         if ( this->astFunc->nbArg() != this->operArgs )
            throw std::runtime_error("args cardinality error");
       }
      else if ( this->astTree->isBoolean() )
       {
         this->noOper    = true;
         this->valNoOper = this->astTree->getBoolean();
       }
      else
         throw std::runtime_error("unexpected AST Predicate");
    }


   gec::IntVar pred::getPowVarFromExpr (cxp::ASTList const      & args,
                                        gec::IntVarArray const  & gVars, 
                                        gec::Space              & gSpace,
                                        cxp::AST const          & expr,
                                        int                       valExp)
    {
      if ( 0 >= valExp )
         throw std::runtime_error("pow() exponent: bad value");

      gec::IntVar  ret;

      if ( 1 == valExp )
         ret = pred::getIntVarFromExpr(args, gVars, gSpace, expr);
      else if ( 0 == (1 & valExp) ) // even case (valExp >= 2)
         ret = gec::expr(
            gSpace,
            gec::sqr(pred::getPowVarFromExpr(args, gVars, gSpace, expr,
                                             (valExp >> 1))));
      else if ( 1 == (1 & valExp) ) // odd case (valExp > 2)
         ret = gec::expr(
            gSpace,
            gec::sqr(pred::getPowVarFromExpr(args, gVars, gSpace, expr,
                                             (valExp >> 1)))
            * pred::getIntVarFromExpr(args, gVars, gSpace, expr));

      return(ret);
    }

   gec::IntVar pred::getIntVarFromExpr (cxp::ASTList const      & args,
                                        gec::IntVarArray const  & gVars, 
                                        gec::Space              & gSpace,
                                        cxp::AST const          & expr)
    {
      gec::IntVar  ret;

      if ( expr.isInteger() )
         ret = constr::getIntVarFromInt(gSpace, expr);
      else if ( expr.isVar() )
         ret = constr::getIntVarFromDVar(gVars, gSpace,
                                         args[expr.getVarId()]);
      else if ( expr.isOperator() )
       {
         cxp::ASTAbstractFunction const  & aF
            = dynamic_cast<cxp::ASTAbstractFunction const &>(expr);

         NodeType  oper(aF.getType());
         int       operArgs(pred::getNArgs(oper));

         if ( false == pred::isIntegerOp(oper) )
            throw std::runtime_error("integer operator failure");

         if ( aF.nbArg() != operArgs )
            throw std::runtime_error("args cardinality failure");

         switch ( oper )
          {
            case F_ABS:
               ret = gec::expr(
                  gSpace,
                  gec::abs(
                     pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(0))));
               break;

            case F_NEG:
               ret = gec::expr(
                  gSpace,
                  - pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(0)));
               break;

            case F_MIN:
               ret = gec::expr(
                  gSpace,
                  gec::min(
                     pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(0)),
                     pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1))));
               break;

            case F_MAX:
               ret = gec::expr(
                  gSpace,
                  gec::max(
                     pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(0)),
                     pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1))));
               break;

            case F_MOD:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  % pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_POW:
                {
                  // n.b.: we, at the moment (?), support only pox(x,y) when
                  // y is assigned (a integer val or a integer expression or
                  // a parameter assigned to a integer constant or to a 
                  // variable with a domain with a only a value)
                  gec::IntVar  iVExp(pred::getIntVarFromExpr(
                        args, gVars, gSpace, aF.getArg(1)));

                  // Warning: don't change the start of the throw
                  // message (must begin with "unsupported " for
                  // a CSC-09 compliant exit message)
                  if ( false == iVExp.assigned() )
                     throw std::runtime_error("unsupported pow() argument");

                  if ( 0 > iVExp.val() )
                     ret = gec::expr(
                        gSpace,
                        1
                        / pred::getPowVarFromExpr(args, gVars, gSpace,
                                                  aF.getArg(0),
                                                  - iVExp.val()));
                  else if ( 0 == iVExp.val() )
                     ret = gec::IntVar(gSpace, 1, 1);
                  else
                     ret = pred::getPowVarFromExpr(args, gVars, gSpace,
                                                   aF.getArg(0),
                                                   iVExp.val());
                 }
               break;

            case F_ADD:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  + pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_SUB:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  - pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_MUL:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  * pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_DIV:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  / pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_IF:
                {
                  gec::BoolVar
                     b(pred::getBoolVarFromExpr(args, gVars, gSpace,
                                                aF.getArg(0)));

                  ret = gec::expr(
                     gSpace,
                     (pred::getIntVarFromExpr(args, gVars, gSpace,
                                              aF.getArg(1))
                      * b)
                     + (pred::getIntVarFromExpr(args, gVars, gSpace,
                                                aF.getArg(2))
                        * gec::expr(gSpace, ! b) ));
                }
               break;

            default:
               throw std::runtime_error("unexpected operator");
               break;
          }
       }
      else
         throw std::runtime_error("unexpected AST integer expression");

      return(ret);
    }


   gec::BoolVar pred::getBoolVarFromExpr (cxp::ASTList const      & args,
                                          gec::IntVarArray const  & gVars, 
                                          gec::Space              & gSpace,
                                          cxp::AST const          & expr)
    {
      gec::BoolVar  ret;

      if ( expr.isBoolean() )
         ret = constr::getBoolVarFromBool(gSpace, expr);
      else if ( expr.isVar() )
         throw std::runtime_error("unexpected boolean variable");
      else if ( expr.isOperator() )
       {
         cxp::ASTAbstractFunction const  & aF
            = dynamic_cast<cxp::ASTAbstractFunction const &>(expr);

         NodeType  oper(aF.getType());
         int       operArgs(pred::getNArgs(oper));

         if ( false == pred::isBooleanOp(oper) )
            throw std::runtime_error("boolean operator failure");

         if ( aF.nbArg() != operArgs )
            throw std::runtime_error("args cardinality failure");

         switch ( oper )
          {
            case F_EQ:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  == pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1)));
               break;

            case F_NE:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  != pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1)));
               break;

            case F_GE:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  >= pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1)));
               break;

            case F_GT:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  > pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_LE:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  <= pred::getIntVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1)));
               break;

            case F_LT:
               ret = gec::expr(
                  gSpace,
                  pred::getIntVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  < pred::getIntVarFromExpr(args, gVars, gSpace,
                                            aF.getArg(1)));
               break;

            case F_AND:
               ret = gec::expr(
                  gSpace,
                  pred::getBoolVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  && pred::getBoolVarFromExpr(args, gVars, gSpace,
                                              aF.getArg(1)));
               break;

            case F_OR:
               ret = gec::expr(
                  gSpace,
                  pred::getBoolVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  || pred::getBoolVarFromExpr(args, gVars, gSpace,
                                              aF.getArg(1)));
               break;

            case F_XOR:
               ret = gec::expr(
                  gSpace,
                  pred::getBoolVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  ^ pred::getBoolVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(1)));
               break;

            case F_NEG:
               ret = gec::expr(
                  gSpace,
                  ! pred::getBoolVarFromExpr(args, gVars, gSpace,
                                             aF.getArg(0)));
               break;

            case F_IFF:
               ret = gec::expr(
                  gSpace,
                  pred::getBoolVarFromExpr(args, gVars, gSpace, aF.getArg(0))
                  == pred::getBoolVarFromExpr(args, gVars, gSpace,
                                              aF.getArg(1)));
               break;

            default:
               throw std::runtime_error("unexpected operator");
               break;
          }
       }
      else
         throw std::runtime_error("unexpected AST boolean expression");

      return(ret);
    }


   void pred::manageInt2 (cxp::ASTList const      & args,
                          gec::IntVarArray const  & gVars,
                          gec::Space              & gSpace) const
    {
      gec::IntVar  exprl(this->getIntVarFromExpr(args, gVars, gSpace,
                                                 this->astFunc->getArg(0)));
      gec::IntVar  exprr(this->getIntVarFromExpr(args, gVars, gSpace,
                                                 this->astFunc->getArg(1)));

      if ( exprr.assigned() )
         gec::rel(gSpace, exprl, this->getRelOp(this->oper), exprr.val());
      else if ( exprl.assigned() )
         gec::rel(gSpace, exprr, this->getRevRelOp(this->oper), exprl.val());
      else
         gec::rel(gSpace, exprl, this->getRelOp(this->oper), exprr);
    }


   void pred::manage (cxp::ASTList const      & args,
                      gec::IntVarArray const  & gVars,
                      gec::Space              & gSpace) const
    {
      if ( this->noOper )
       {
         gec::rel(
            gSpace,
            gec::BoolVar(gSpace, this->valNoOper, this->valNoOper),
            gec::IRT_EQ,
            this->getRelOp(this->oper),
            gec::BoolVar(gSpace, 1, 1));
       }
      else if ( F_NOT == this->oper )
         this->manageNot(args, gVars, gSpace);
      else if ( this->isBooleanOp2BoolArgs(this->oper) )
         this->manageBool2(args, gVars, gSpace);
      else if ( this->isBooleanOp2IntArgs(this->oper) )
         this->manageInt2(args, gVars, gSpace);
      else   
         throw std::runtime_error("unexpected boolean expression");
    }

   std::map<NodeType, int>  pred::initMOpA ()
    {
      std::map<NodeType, int>  ret;

      ret[F_NEG] = ret[F_ABS] = ret[F_NOT] =                 1;
      ret[F_ADD] = ret[F_SUB] = ret[F_MUL] = ret[F_DIV] =
         ret[F_MOD] = ret[F_POW] = ret[F_MIN] = ret[F_MAX] =
         ret[F_EQ]  = ret[F_NE]  = ret[F_GE]  = ret[F_GT]  =
         ret[F_LE]  = ret[F_LT]  = ret[F_AND] = ret[F_OR]  =
         ret[F_XOR] = ret[F_IFF] =                           2;
      ret[F_IF] =                                            3;

      return(ret);
    }


   std::map<NodeType, int> const  pred::mOpA(pred::initMOpA());


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   // - Definition:   
   //    "alldifferent" := "Enforce all variables of the collection 𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂
   //                       to take distinct values"
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂, 𝚟𝚊𝚛) [xcsp implicit]
   // ***********************************************************************
   void globAllDifferent::manage (cxp::ASTList const      & args,
                                  gec::IntVarArray const  & gVars,
                                  gec::Space              & gSpace) const
    {
      // problem: in the "allDifferent" case, arguments can be directly in
      // 'args' (old deprecated syntax?) or in the ASTList in 'args[0]' (new
      // syntax?); we manage this with the following hack
      cxp::ASTList const & aList(
         dynamic_cast<cxp::ASTList const &>(
            ((1==args.nbArg()) && args[0].isList()) ? args[0] : args ));

      gec::IntVarArgs  diffVars;

      for ( int i=0 ; i < aList.nbArg() ; ++i )
         diffVars << this->getIntVarFromDVar(gVars, gSpace, aList[i]);

      gec::distinct(gSpace, diffVars);
    }

   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝙽𝚅𝙰𝚁       𝚍𝚟𝚊𝚛
   //    𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   //    𝚅𝙰𝙻𝚄𝙴𝚂     𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚕−𝚒𝚗𝚝)
   // - Definition:   
   //    "among" := "𝙽𝚅𝙰𝚁 is the number of variables of the collection
   //                𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂 that take their value in 𝚅𝙰𝙻𝚄𝙴𝚂"
   // - Restrictions:
   //    (a) 𝙽𝚅𝙰𝚁 >= 0
   //    (b) 𝙽𝚅𝙰𝚁 <= |𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂|
   //    (c) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂, 𝚟𝚊𝚛) [xcsp implicit]
   //    (d) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝙻𝚄𝙴𝚂, 𝚟𝚊𝚕)    [xcsp implicit]
   //    (e) 𝚍𝚒𝚜𝚝𝚒𝚗𝚌𝚝(𝚅𝙰𝙻𝚄𝙴𝚂, 𝚟𝚊𝚕)
   // ***********************************************************************
   void globAmong::manage (cxp::ASTList const      & args,
                           gec::IntVarArray const  & gVars,
                           gec::Space              & gSpace) const
    {
      if ( 3 != args.nbArg() )
         throw std::runtime_error("among args cardinality error");

      cxp::AST const  & nVar(args[0]);
      cxp::AST const  & vars(args[1]);
      cxp::AST const  & vals(args[2]);

      if ( false == vars.isList() ) 
         throw std::runtime_error("among arg 2 error");

      if ( false == vals.isList() ) 
         throw std::runtime_error("among arg 3 error");

      gec::IntVar  nV = this->getIntVarFromDVar(gVars, gSpace, nVar);

      std::vector<int>  tmpV(vals.nbArg());

      for ( int i=0 ; i < vals.nbArg() ; ++i )
         tmpV[i] = vals[i].getInteger();

      gec::IntSet    iSVals(tmpV.data(), tmpV.size());

      // apparently we can use gec::sequence, but this function throw an
      // exception when the second argument "contain the same unassigned 
      // variable multiply"; this ins't a requirement in "among"
      // gec::sequence(gSpace, iVAVars, iSVals, iVAVars.size(), nV, nV);

      // restriction (e)
      if ( int(iSVals.size()) != vals.nbArg() )
         throw std::runtime_error("among restriction (e) error");

      // restrictions (a) & (b)
      gec::rel(gSpace, nV, gec::IRT_GQ, 0);
      gec::rel(gSpace, nV, gec::IRT_LQ, vals.nbArg());

      gec::IntVar  gEx(gSpace, 0, 0);

      for ( int i=0 ; i < vars.nbArg() ; ++i )
         gEx = gec::expr(
            gSpace,
            gEx
            + gec::expr(
               gSpace,
               gec::singleton(this->getIntVarFromDVar(gVars, gSpace,
                                                      vars[i]))
               <= iSVals));

      // among contraint
      gec::rel(gSpace, gEx, gec::IRT_EQ, nV);
    }


   // ***********************************************************************
   // Global Constraint Catalogue:
   // - Arguments:
   //    𝙽          𝚒𝚗𝚝
   //    𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   //    𝚅𝙰𝙻𝚄𝙴      𝚒𝚗𝚝
   // - Definition:   
   //    "atleast" := "At least 𝙽 variables of the 𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂 collection are
   //                  assigned value 𝚅𝙰𝙻𝚄𝙴"
   // - Restrictions:
   //    (a) 𝙽 >= 0
   //    (b) 𝙽 <= |𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂|
   //    (c) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂, 𝚟𝚊𝚛) [xcsp implicit]
   // ***********************************************************************
   // ***********************************************************************
   // Global Constraint Catalogue:
   // - Arguments:
   //    𝙽 𝚒𝚗𝚝
   //    𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂 𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   //    𝚅𝙰𝙻𝚄𝙴 𝚒𝚗𝚝
   // - Definition:   
   //    "atmost" = "At most 𝙽 variables of the 𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂 collection are
   //                assigned value 𝚅𝙰𝙻𝚄𝙴.
   // - Restrictions:
   //    (a) 𝙽 >= 0
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂, 𝚟𝚊𝚛) [xcsp implicit]
   // ***********************************************************************
   void globAtBase::verifAt (cxp::ASTList const      & args,
                             gec::IntVarArray const  & gVars,
                             gec::Space              & gSpace,
                             gec::IntRelType           gRel)
    {
      if ( 3 != args.nbArg() )
         throw std::runtime_error("verifAt args cardinality error");

      cxp::AST const  & n(args[0]);
      cxp::AST const  & vars(args[1]);
      cxp::AST const  & val(args[2]);

      if ( false == n.isInteger() ) 
         throw std::runtime_error("verifAt arg 1 error");

      if ( false == vars.isList() ) 
         throw std::runtime_error("verifAt arg 2 error");

      if ( false == val.isInteger() ) 
         throw std::runtime_error("verifAt arg 3 error");

      int  nVal(n.getInteger());
      int  valVal(val.getInteger());

      // restriction atleat (a) && atmost (a)
      if ( 0 > nVal )
         throw std::runtime_error("verifAt restriction (a)/(a) error");

      // restriction atleat (b) [atleast only]
      if ( (gec::IRT_GQ == gRel) && (vars.nbArg() < nVal) )
         throw std::runtime_error("verifAt restriction (b)/() error");
      
      gec::IntVarArgs  iVAVars;

      for ( int i=0 ; i < vars.nbArg() ; ++i )
         iVAVars << constr::getIntVarFromDVar(gVars, gSpace, vars[i]);

      // constraint (gRel is gec::IRT_GQ for "atleast", gec::IRT_LQ for
      // "atmost")
      gec::count(gSpace, iVAVars, valVal, gRel, nVal);
    }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚃𝙰𝚂𝙺𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚘𝚛𝚒𝚐𝚒𝚗−𝚍𝚟𝚊𝚛, 𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗−𝚍𝚟𝚊𝚛, 𝚎𝚗𝚍−𝚍𝚟𝚊𝚛,
   //                      𝚑𝚎𝚒𝚐𝚑𝚝−𝚍𝚟𝚊𝚛)
   //    𝙻𝙸𝙼𝙸𝚃  𝚒𝚗𝚝
   // - Definition:   
   //    "cumulative" := "Consider a set T of tasks described by the 𝚃𝙰𝚂𝙺𝚂
   //                     collection. The 𝚌𝚞𝚖𝚞𝚕𝚊𝚝𝚒𝚟𝚎 constraint enforces that
   //                     at each point in time, the cumulated height of the
   //                     set of tasks that overlap that point, does not
   //                     exceed a given limit. A task overlaps a point i if
   //                     and only if (1) its origin is less than or equal to
   //                     i, and (2) its end is strictly greater than i. It
   //                     also imposes for each task of T the constraint
   //                     𝚘𝚛𝚒𝚐𝚒𝚗+𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗=𝚎𝚗𝚍."
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎_𝚊𝚝_𝚕𝚎𝚊𝚜𝚝(2, 𝚃𝙰𝚂𝙺𝚂, [𝚘𝚛𝚒𝚐𝚒𝚗, 𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗, 𝚎𝚗𝚍])
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚃𝙰𝚂𝙺𝚂, 𝚑𝚎𝚒𝚐𝚑𝚝)
   //    (c) 𝚃𝙰𝚂𝙺𝚂.𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗 >= 0
   //    (d) 𝚃𝙰𝚂𝙺𝚂.𝚘𝚛𝚒𝚐𝚒𝚗 <= 𝚃𝙰𝚂𝙺𝚂.𝚎𝚗𝚍
   //    (e) 𝚃𝙰𝚂𝙺𝚂.𝚑𝚎𝚒𝚐𝚑𝚝 >= 0
   //    (f) 𝙻𝙸𝙼𝙸𝚃 >= 0
   // - Remarks:
   //    - "Some systems like Ilog CP Optimizer also assume that a
   //       zero-duration task overlaps a point i if and only if (1) its
   //       origin is less than or equal to i, and (2) its end is greater
   //       than or equal to i. Under this definition, the height of a
   //       zero-duration task is also taken into account in the resource
   //       consumption profile."
   // ***********************************************************************
   void globCumulative::manage (cxp::ASTList const      & args,
                                gec::IntVarArray const  & gVars,
                                gec::Space              & gSpace) const
    {
      if ( 2 != args.nbArg() )
         throw std::runtime_error("cumulative args cardinality error");

      cxp::AST const  & tasks(args[0]);
      cxp::AST const  & limit(args[1]);

      if ( false == tasks.isList() ) 
         throw std::runtime_error("cumulative arg 1 error");

      if ( false == limit.isInteger() ) 
         throw std::runtime_error("cumulative arg 2 error");

      gec::IntVarArgs  iVAOrigin;
      gec::IntVarArgs  iVADuration;
      gec::IntVarArgs  iVAEnd;
      gec::IntVarArgs  iVAHeight;

      int const numTsk(tasks.nbArg());
      int const limVal(limit.getInteger());

      enum { noOrigin, noDuration, noEnd, full}  hasCase;

      // restriction (f)
      if ( 0 > limVal )
         throw std::runtime_error("cumulative restriction (f) error");

      for ( int i=0 ; i < numTsk ; ++i )
       {
         cxp::ASTDict const  & aD
            = dynamic_cast<cxp::ASTDict const &>(tasks[i]);

         // XCSP 2.1 specifications: "one attribute among {origin, duration,
         // end} may be missing" (pag. 27), so...
         if ( false == aD.hasKey("origin") ) 
          {
            hasCase = noOrigin;

            // no "origin", so "duration" and "end" are mandatory and
            // "origin := end - duration" (restriction (a) satisfied or
            // exception)
            iVADuration
               << this->getIntVarFromDVar(gVars, gSpace, aD["duration"]);
            iVAEnd
               << this->getIntVarFromDVar(gVars, gSpace, aD["end"]);
            iVAOrigin
               << gec::expr(gSpace, iVAEnd[i] - iVADuration[i]);
          }
         else if ( false == aD.hasKey("duration") ) 
          {
            hasCase = noDuration;

            // no "duration", so "origin" and "end" are mandatory and
            // "duration := end - origin" (restriction (a) satisfied or
            // exception)
            iVAOrigin
               << this->getIntVarFromDVar(gVars, gSpace, aD["origin"]);
            iVAEnd
               << this->getIntVarFromDVar(gVars, gSpace, aD["end"]);
            iVADuration
               << gec::expr(gSpace, iVAEnd[i] - iVAOrigin[i]);
          }
         else if ( false == aD.hasKey("end") ) 
          {
            hasCase = noEnd;

            // no "end", so "origin" and "duration" are mandatory and
            // "end := origin + duration" (restriction (a) satisfied or
            // exception)
            iVAOrigin
               << this->getIntVarFromDVar(gVars, gSpace, aD["origin"]);
            iVADuration
               << this->getIntVarFromDVar(gVars, gSpace, aD["duration"]);
            iVAEnd
               << gec::expr(gSpace, iVAOrigin[i] + iVADuration[i]);
          }
         else
          {
            hasCase = full;

            // we have "origin", "duration" and "end"; we must impose
            // the constraint "end = origin + duration" (Gecode don't
            // test this in comulative() and comulatives()) (restriction
            // (a) satisfied)
            iVAOrigin
               << this->getIntVarFromDVar(gVars, gSpace, aD["origin"]);
            iVADuration
               << this->getIntVarFromDVar(gVars, gSpace, aD["duration"]);
            iVAEnd
               << this->getIntVarFromDVar(gVars, gSpace, aD["end"]);

            gec::rel(
               gSpace,
               gec::expr(gSpace, iVAOrigin[i] + iVADuration[i]),
               gec::IRT_EQ,
               iVAEnd[i]);
          }

         // "height" is mandatory (restriction (b) satisfied or exception)
         iVAHeight
            << this->getIntVarFromDVar(gVars, gSpace, aD["height"]);

         // for each task we have (and we must, according to the case,
         // impose) the following restrictions:
         // (c) duration >= 0 (all cases)
         // (e) height >= 0   (all cases)
         // (d) origin <= end (only case full; superflous otherwise (obvious
         //     if origin or end is calculated (with duration >=0); derive
         //     from (c) when duration in calculated))
         gec::rel(gSpace, iVADuration[i], gec::IRT_GQ, 0);
         gec::rel(gSpace, iVAOrigin[i],   gec::IRT_LQ, iVAEnd[i]);

         if ( full == hasCase )
            gec::rel(gSpace, iVAHeight[i], gec::IRT_GQ, 0);
       }

      // 2 cases:
      // (1) all heights are assigned -> we can transform iVAHeight in a
      //     IntArgs and call cumulatives()
      // (2) otherwise (there is at most a true variable in iVAHeight) ->
      //     we can only use cumulatives()
      if ( iVAHeight.assigned() )
       {
         gec::IntArgs  iAHeight(numTsk);

         for ( int i = 0 ; i < numTsk ; ++i )
            iAHeight[i] = iVAHeight[i].val();

         gec::cumulative(gSpace, limVal, iVAOrigin, iVADuration, iVAEnd,
                         iAHeight);
       }
      else
         gec::cumulatives(gSpace, gec::IntArgs::create(numTsk, 0, 0),
                          iVAOrigin, iVADuration, iVAEnd, iVAHeight,
                          gec::IntArgs(1, limVal), true);
    }


   void globDiffn::setDiffn4two (gec::IntVarArray const  & gVars,
                                 gec::Space              & gSpace,
                                 globDiffn::orth         & o1,
                                 globDiffn::orth         & o2)
    {
      gec::BoolVar  bVal(gSpace, 0, 0);

      for ( unsigned int ui = 0 ; ui < o1.size() ; ++ui )
         bVal = gec::expr(
            gSpace,
            bVal
            || (o1[ui][globDiffn::oEnd] <= o2[ui][globDiffn::oOri])
            || (o2[ui][globDiffn::oEnd] <= o1[ui][globDiffn::oOri]) );

      gec::rel(gSpace, bVal, gec::IRT_EQ, 1);
    }


   // - residual Restrictions:
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎_𝚊𝚝_𝚕𝚎𝚊𝚜𝚝(2, 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴, [𝚘𝚛𝚒, 𝚜𝚒𝚣, 𝚎𝚗𝚍])
   //    (c) 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚜𝚒𝚣 >= 0
   //    (d) 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚘𝚛𝚒 <= 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚎𝚗𝚍
   void globDiffn::loadTriple (gec::IntVarArray const  & gVars,
                               gec::Space              & gSpace,
                               globDiffn::orthTriple   & gdfT,
                               cxp::AST const          & astT)
    {
      cxp::ASTDict const &  aD(dynamic_cast<cxp::ASTDict const &>(astT));

      // 4 cases:
      // (1) no "ori"
      // (2) no "siz"
      // (3) no "end"
      // (4) "ori", "siz" and "and"
      if ( false == aD.hasKey("ori") ) 
       {
         // case (1): no "ori", so "siz" and "end" are mandatory;
         // restriction (b) satisfied or exception
         gdfT[globDiffn::oSiz]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["siz"]);
         gdfT[globDiffn::oEnd]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["end"]);

         // now we can calculate "ori"
         gdfT[globDiffn::oOri]
            = gec::expr(gSpace,
                        gdfT[globDiffn::oEnd] - gdfT[globDiffn::oSiz]);
       }
      else if ( false == aD.hasKey("siz") ) 
       {
         // case (2): no "siz", so "ori" and "end" are mandatory;
         // restriction (b) satisfied or exception
         gdfT[globDiffn::oOri]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["ori"]);
         gdfT[globDiffn::oEnd]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["end"]);

         // now we can calculate "siz"
         gdfT[globDiffn::oSiz]
            = gec::expr(gSpace,
                        gdfT[globDiffn::oEnd] - gdfT[globDiffn::oOri]);
       }
      else if ( false == aD.hasKey("end") ) 
       {
         // case (3): no "end", so "ori" and "siz" are mandatory;
         // restriction (b) satisfied or exception
         gdfT[globDiffn::oOri]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["ori"]);
         gdfT[globDiffn::oSiz]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["siz"]);

         // now we can calculate "end"
         gdfT[globDiffn::oEnd]
            = gec::expr(gSpace,
                        gdfT[globDiffn::oOri] + gdfT[globDiffn::oSiz]);
       }
      else
       {
         // case (4): so "ori", "siz" and "and"; restriction (b) satisfied
         gdfT[globDiffn::oOri]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["ori"]);
         gdfT[globDiffn::oSiz]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["siz"]);
         gdfT[globDiffn::oEnd]
            = constr::getIntVarFromDVar(gVars, gSpace, aD["end"]);

         // we must (?) impose (only in this case) that ori+siz=end
         gec::rel(
            gSpace,
            gec::expr(gSpace, gdfT[globDiffn::oOri] + gdfT[globDiffn::oSiz]),
            gec::IRT_EQ,
            gdfT[globDiffn::oEnd]);
       }

      // now we can impose (c) and (d) restrictions
      gec::rel(gSpace, gdfT[globDiffn::oSiz], gec::IRT_GQ, 0);
      gec::rel(gSpace, gdfT[globDiffn::oOri], gec::IRT_LQ,
               gdfT[globDiffn::oEnd]);
    }


   // - residual Restrictions:
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎_𝚊𝚝_𝚕𝚎𝚊𝚜𝚝(2, 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴, [𝚘𝚛𝚒, 𝚜𝚒𝚣, 𝚎𝚗𝚍])
   //    (c) 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚜𝚒𝚣 >= 0
   //    (d) 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚘𝚛𝚒 <= 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚎𝚗𝚍
   //    (f) 𝚜𝚊𝚖𝚎_𝚜𝚒𝚣𝚎(𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴𝚂, 𝚘𝚛𝚝𝚑)
   void globDiffn::loadOrthotope (gec::IntVarArray const  & gVars,
                                  gec::Space              & gSpace,
                                  globDiffn::orth         & gdfO,
                                  cxp::AST const          & astO,
                                  int                       numTpl)
    {
      cxp::ASTList const  & aL(dynamic_cast<cxp::ASTList const &>(astO));

      // now we set the numTpl value (loading the first orthotope)
      // or we check the restriction (f)
      if ( 0 == numTpl )
         numTpl = aL.nbArg();
      else if ( aL.nbArg() != numTpl )
         throw std::runtime_error("diffn dimensional error");

      gdfO.resize(numTpl);

      // we demand to loadTriple the check of residual (b), (c) and (d)
      for ( int i = 0 ; i < aL.nbArg() ; ++i )
         globDiffn::loadTriple(gVars, gSpace, gdfO[i], aL[i]);
    }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    ORTHOTOPES  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(orth−ORTHOTOPE)
   //    [where ORTHOTOPE  collection(𝚘𝚛𝚒−𝚍𝚟𝚊𝚛, 𝚜𝚒𝚣−𝚍𝚟𝚊𝚛, 𝚎𝚗𝚍−𝚍𝚟𝚊𝚛)]
   // - Definition:
   //    "diffn" := "Generalised multi-dimensional non-overlapping
   //                constraint: Holds if, for each pair of orthotopes
   //                (O1,O2), O1 and O2 do not overlap. Two orthotopes do not
   //                overlap if there exists at least one dimension where
   //                their projections do not overlap"
   // - Restrictions:
   //    (a) |𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴| > 0
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎_𝚊𝚝_𝚕𝚎𝚊𝚜𝚝(2, 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴, [𝚘𝚛𝚒, 𝚜𝚒𝚣, 𝚎𝚗𝚍])
   //    (c) 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚜𝚒𝚣 >= 0
   //    (d) 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚘𝚛𝚒 <= 𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴.𝚎𝚗𝚍
   //    (e) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴𝚂, 𝚘𝚛𝚝𝚑)                      [xcsp implicit]
   //    (f) 𝚜𝚊𝚖𝚎_𝚜𝚒𝚣𝚎(𝙾𝚁𝚃𝙷𝙾𝚃𝙾𝙿𝙴𝚂, 𝚘𝚛𝚝𝚑)
   // ***********************************************************************
   void globDiffn::manage (cxp::ASTList const      & args,
                           gec::IntVarArray const  & gVars,
                           gec::Space              & gSpace) const
    {
      if ( 1 != args.nbArg() )
         throw std::runtime_error("diffn args cardinality error");

      cxp::AST const  & orths(args[0]);

      if ( false == orths.isList() ) 
         throw std::runtime_error("diffn arg 1 error");

      int  nOrths(orths.nbArg());

      // restriction (a)
      if ( 0 >= nOrths ) 
         throw std::runtime_error("diffn restriction (a) error");

      std::vector<globDiffn::orth>  oV(nOrths);

      // orthotope loading with check or imposition of restrictions (b),
      // (c), (d) and (from the second) (f)

      // we load the first orthotope, so we can determine the dimension
      // (the number of the triples) for all the orthotopes 
      globDiffn::loadOrthotope(gVars, gSpace, oV[0], orths[0], 0);

      int  numTpl(int((oV[0]).size()));

      // determined the dimentin of the first orthotope, we can load the
      // others checking the dimension (restriction (f))
      for ( int i = 1 ; i < nOrths ; ++i )
         globDiffn::loadOrthotope(gVars, gSpace, oV[i], orths[i], numTpl);

      // now, for every couple of orthotopes, we can impose the contraint
      for ( int i = 0U ; i < (nOrths - 1) ; ++i )
         for ( int j = i + 1U ; j < nOrths ; ++j )
            globDiffn::setDiffn4two(gVars, gSpace, oV[i], oV[j]);
    }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚃𝙰𝚂𝙺𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚘𝚛𝚒𝚐𝚒𝚗−𝚍𝚟𝚊𝚛, 𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗−𝚍𝚟𝚊𝚛)
   // - Definition:   
   //    "disjunctive" := "All the tasks of the collection 𝚃𝙰𝚂𝙺𝚂 that have a
   //                      duration strictly greater than 0 should not
   //                      overlap"
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚃𝙰𝚂𝙺𝚂, [𝚘𝚛𝚒𝚐𝚒𝚗, 𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗])
   //    (b) 𝚃𝙰𝚂𝙺𝚂.𝚍𝚞𝚛𝚊𝚝𝚒𝚘𝚗 >= 0
   // - Remarks:
   //    - "All the tasks of the collection 𝚃𝙰𝚂𝙺𝚂 that have a duration
   //       strictly greater than 0 should not overlap"
   //    - "Some systems like Ilog CP Optimizer also imposes that zero
   //       duration tasks do not overlap non -zero duration tasks"
   // ***********************************************************************
   void globDisjunctive::manage (cxp::ASTList const      & args,
                                 gec::IntVarArray const  & gVars,
                                 gec::Space              & gSpace) const
    {
      if ( 1 != args.nbArg() )
         throw std::runtime_error("disjunctive args cardinality error");

      cxp::AST const  & tasks(args[0]);

      if ( false == tasks.isList() ) 
         throw std::runtime_error("disjunctive arg 1 error");

      gec::IntVarArgs  iVAOrigin;
      gec::IntVarArgs  iVADuration;

      int const numTsk(tasks.nbArg());

      for ( int i=0 ; i < numTsk ; ++i )
       {
         cxp::ASTDict const  & aD
            = dynamic_cast<cxp::ASTDict const &>(tasks[i]);

         // restriction (a) satisfied or exception
         iVAOrigin
            << this->getIntVarFromDVar(gVars, gSpace, aD["origin"]);
         iVADuration
            << this->getIntVarFromDVar(gVars, gSpace, aD["duration"]);

         // restriction (b)
         gec::rel(gSpace, iVADuration[i], gec::IRT_GQ, 0);
       }

      // 2 cases:
      // (1) all duration are assigned -> we can transform iVADuration in a
      //     IntArgs and call the apposite version of unary()
      // (2) otherwise (there is at most a true variable in iVAHeight) ->
      //     we must use the version of unary that receive the origin,
      //     duration and end IntVarArgs, so we must generate the end 
      //     IntVarArgs
      if ( iVADuration.assigned() )
       {
         gec::IntArgs  iADuration(numTsk);

         for ( int i = 0 ; i < numTsk ; ++i )
            iADuration[i] = iVADuration[i].val();

         gec::unary(gSpace, iVAOrigin, iADuration);
       }
      else
       {
         gec::IntVarArgs  iVAEnd;

         for ( int i = 0 ; i < numTsk ; ++i )
            iVAEnd << gec::expr(gSpace, iVAOrigin[i] + iVADuration[i]);

         gec::unary(gSpace, iVAOrigin, iVADuration, iVAEnd);
       }
    }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝙸𝙽𝙳𝙴𝚇  𝚍𝚟𝚊𝚛
   //    𝚃𝙰𝙱𝙻𝙴  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚕𝚞𝚎−𝚍𝚟𝚊𝚛)
   //    𝚅𝙰𝙻𝚄𝙴  𝚍𝚟𝚊𝚛
   // - Definition:
   //    "element" := "𝚅𝙰𝙻𝚄𝙴 is equal to the 𝙸𝙽𝙳𝙴𝚇th item of 𝚃𝙰𝙱𝙻𝙴"
   // - Restrictions:
   //    (a) 𝙸𝙽𝙳𝙴𝚇 >= 1             [gecoded (INDEX:=INDEX-1): INDEX >= 0]
   //    (b) 𝙸𝙽𝙳𝙴𝚇 <= |𝚃𝙰𝙱𝙻𝙴|       [gecoded (INDEX:=INDEX-1):
   //                                  INDEX < |TABLE|]
   //    (c) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚃𝙰𝙱𝙻𝙴, 𝚟𝚊𝚕𝚞𝚎) [xcsp implicit]
   // ***********************************************************************
   void globElement::manage (cxp::ASTList const      & args,
                             gec::IntVarArray const  & gVars,
                             gec::Space              & gSpace) const
    {
      if ( 3 != args.nbArg() )
         throw std::runtime_error("element args cardinality error");

      cxp::AST const  & index(args[0]);
      cxp::AST const  & list(args[1]);
      cxp::AST const  & value(args[2]);

      if ( false == list.isList() ) 
         throw std::runtime_error("element arg 2 error");

      gec::IntVar  vInd(this->getIntVarFromDVar(gVars, gSpace, index));
      gec::IntVar  vVal(this->getIntVarFromDVar(gVars, gSpace, value));

      // n.b.: the index value in Gecode start from zero ([0..n-1]);
      // in XCSP start from 1 ([1..n]); we must correct (subtract 1
      // from) the index
      if ( vInd.assigned() )
       {
         int  gIdx(vInd.val()-1);

         // restrictions (a) and (b)
         if ( (0 > gIdx) || (list.nbArg() <= gIdx) )
            throw std::runtime_error("element idx 1 error");

         // constraint (case (2))
         if ( vVal.assigned() )
            gec::rel(
               gSpace,
               this->getIntVarFromDVar(gVars, gSpace, list[gIdx]),
               gec::IRT_EQ,
               vVal.val());
         else
            gec::rel(
               gSpace,
               this->getIntVarFromDVar(gVars, gSpace, list[gIdx]),
               gec::IRT_EQ,
               vVal);
       }
      else
       {
         gec::IntVarArgs  iVA;

         for ( int i=0 ; i < list.nbArg() ; ++i )
            iVA << this->getIntVarFromDVar(gVars, gSpace, list[i]);

         gec::IntVar  vIndM1(gec::expr(gSpace, vInd-1));

         // restrictions (a) && (b)
         gec::rel(gSpace, vIndM1, gec::IRT_GQ, 0);
         gec::rel(gSpace, vIndM1, gec::IRT_LE, iVA.size());

         // constraint (case (1))
         if ( vVal.assigned() )
            gec::element(gSpace, iVA, vIndM1, vVal.val());
         else
            gec::element(gSpace, iVA, vIndM1, vVal);
       }
    }


   gec::IntVar globGlobalCardinality::getGCExpr
      (gec::Space             & gSpace,
       gec::IntVarArgs const  & iVA,
       int                      val)
       {
         gec::IntVar  ret = gec::IntVar(gSpace, 0, 0);

         for ( int i = 0 ; i < iVA.size() ; ++i )
            ret = gec::expr(
               gSpace,
               ret + gec::expr(gSpace, (iVA[i] == val)));

         return(ret);
       }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   //    𝚅𝙰𝙻𝚄𝙴𝚂     𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚕−𝚒𝚗𝚝, 𝚗𝚘𝚌𝚌𝚞𝚛𝚛𝚎𝚗𝚌𝚎−𝚍𝚟𝚊𝚛)
   // - Definition:
   //    "global_cardinality" := "Each value 𝚅𝙰𝙻𝚄𝙴𝚂[i].𝚟𝚊𝚕 (1<=i<=|𝚅𝙰𝙻𝚄𝙴𝚂|)
   //                             should be taken by exactly
   //                             𝚅𝙰𝙻𝚄𝙴𝚂[i].𝚗𝚘𝚌𝚌𝚞𝚛𝚛𝚎𝚗𝚌𝚎 variables of the
   //                             𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂 collection"
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂, 𝚟𝚊𝚛)             [xcsp implicit]
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝙻𝚄𝙴𝚂, [𝚟𝚊𝚕, 𝚗𝚘𝚌𝚌𝚞𝚛𝚛𝚎𝚗𝚌𝚎])
   //    (c) 𝚍𝚒𝚜𝚝𝚒𝚗𝚌𝚝(𝚅𝙰𝙻𝚄𝙴𝚂, 𝚟𝚊𝚕)
   //    (d) 𝚅𝙰𝙻𝚄𝙴𝚂.𝚗𝚘𝚌𝚌𝚞𝚛𝚛𝚎𝚗𝚌𝚎 >= 0
   //    (e) 𝚅𝙰𝙻𝚄𝙴𝚂.𝚗𝚘𝚌𝚌𝚞𝚛𝚛𝚎𝚗𝚌𝚎 <= |𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂|
   // ***********************************************************************
   void globGlobalCardinality::manage (cxp::ASTList const      & args,
                                       gec::IntVarArray const  & gVars,
                                       gec::Space              & gSpace) const
    {
      if ( 2 != args.nbArg() )
         throw std::runtime_error("globCard args cardinality error");

      cxp::AST const  & vars(args[0]);
      cxp::AST const  & vals(args[1]);

      if ( false == vars.isList() ) 
         throw std::runtime_error("globCars arg 1 error");

      if ( false == vals.isList() ) 
         throw std::runtime_error("globCars arg 2 error");

      gec::IntVarArgs  iVAVars;

      for ( int i=0 ; i < vars.nbArg() ; ++i )
         iVAVars << this->getIntVarFromDVar(gVars, gSpace, vars[i]);

      std::set<int>  sVals;

      for ( int i=0 ; i < vals.nbArg() ; ++i )
       {
         // restriction (b) satisfied (part "val") or exception
         if ( false == vals[i]["val"].isInteger() )
            throw std::runtime_error("globCars AST type error");

         int  val(vals[i]["val"].getInteger());

         // restriction (c) satisfied or exception
         if ( false == sVals.insert(val).second )
            throw std::runtime_error("globCars non unique val");

         // restriction (b) satisfied (part "noccurrence") or exception
         gec::IntVar nOcc(this->getIntVarFromDVar(gVars, gSpace,
                                                  vals[i]["noccurrence"]));

         // restrictions (d) && (e)
         gec::rel(gSpace, nOcc, gec::IRT_GQ, 0);
         gec::rel(gSpace, nOcc, gec::IRT_LQ, vars.nbArg());

         // constraint
         gec::rel(gSpace, this->getGCExpr(gSpace, iVAVars, val), gec::IRT_EQ,
                  nOcc);
       }

      // we can't use gec::count() because it require that the union of the
      // values in iVAVars is equal to the union of the values in iAVals
      // (that isn't required in this constraint)
      //// gec::count(gSpace, iVAVars, iVAOccurs, iAVals);
    }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚅𝙴𝙲𝚃𝙾𝚁1  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   //    𝚅𝙴𝙲𝚃𝙾𝚁2  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   // - Definition:
   //    "lex_less" := "𝚅𝙴𝙲𝚃𝙾𝚁1 is lexicographically strictly less than
   //                   𝚅𝙴𝙲𝚃𝙾𝚁2. Given two vectors, X and Y of n components,
   //                   <X0,...,Xn−1> and <Y0,...,Yn−1>, X is
   //                   lexicographically strictly less than Y if and only if
   //                   X0<Y0 or X0=Y0 and <X1,...,Xn−1> is lexicographically
   //                   strictly less than <Y1,...,Yn−1>"
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙴𝙲𝚃𝙾𝚁1, 𝚟𝚊𝚛)  [xcsp implicit]
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙴𝙲𝚃𝙾𝚁2, 𝚟𝚊𝚛)  [xcsp implicit]
   //    (c) |𝚅𝙴𝙲𝚃𝙾𝚁1| = |𝚅𝙴𝙲𝚃𝙾𝚁2|
   // ***********************************************************************
   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚅𝙴𝙲𝚃𝙾𝚁1  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   //    𝚅𝙴𝙲𝚃𝙾𝚁2  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   // - Definition:
   //    "lex_lesseq" := "𝚅𝙴𝙲𝚃𝙾𝚁1 is lexicographically less than or equal to
   //                     𝚅𝙴𝙲𝚃𝙾𝚁2. Given two vectors, X and Y of n
   //                     components, <X0,...,Xn−1> and <Y0,...,Yn−1>, X is
   //                     lexicographically less than or equal to Y if and
   //                     only if n=0 or X0<Y0 or X0=Y0 and <X1,...,Xn−1> is
   //                     lexicographically less than or equal to
   //                     <Y1,...,Yn−1>"
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙴𝙲𝚃𝙾𝚁1, 𝚟𝚊𝚛)  [xcsp implicit]
   //    (b) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙴𝙲𝚃𝙾𝚁2, 𝚟𝚊𝚛)  [xcsp implicit]
   //    (c) |𝚅𝙴𝙲𝚃𝙾𝚁1| = |𝚅𝙴𝙲𝚃𝙾𝚁2|
   // ***********************************************************************
   void globLex::setLex (cxp::ASTList const      & args,
                         gec::IntVarArray const  & gVars,
                         gec::Space              & gSpace,
                         gec::IntRelType           gOp) const
    {
      if ( 2 != args.nbArg() )
         throw std::runtime_error("lex args cardinality error");

      cxp::AST const  & v1(args[0]);
      cxp::AST const  & v2(args[1]);

      if ( false == v1.isList() ) 
         throw std::runtime_error("lex arg 1 error");

      if ( false == v2.isList() ) 
         throw std::runtime_error("lex arg 2 error");

      // restriction (c)
      if ( v1.nbArg() != v2.nbArg() ) 
         throw std::runtime_error("lex restriction (c) error");

      gec::IntVarArgs  iVA1;
      gec::IntVarArgs  iVA2;

      for ( int i=0 ; i < v1.nbArg() ; ++i )
       {
         iVA1 << this->getIntVarFromDVar(gVars, gSpace, v1[i]);
         iVA2 << this->getIntVarFromDVar(gVars, gSpace, v2[i]);
       }

      gec::rel(gSpace, iVA1, gOp, iVA2);
    }


   // ***********************************************************************
   // --- Global Constraint Catalogue:
   // - Arguments:
   //    𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(𝚟𝚊𝚛−𝚍𝚟𝚊𝚛)
   // - Definition:
   //    "not_all_equal" := "The variables of the collection 𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂 should
   //                        take more than one single value"
   // - Restrictions:
   //    (a) 𝚛𝚎𝚚𝚞𝚒𝚛𝚎𝚍(𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂, 𝚟𝚊𝚛) [xcsp implicit]
   //    (b) |𝚅𝙰𝚁𝙸𝙰𝙱𝙻𝙴𝚂| > 1
   // ***********************************************************************
   void globNotAllEqual::manage (cxp::ASTList const      & args,
                                 gec::IntVarArray const  & gVars,
                                 gec::Space              & gSpace) const
    {
      if ( 1 != args.nbArg() )
         throw std::runtime_error("notAllEqual args cardinality error");

      cxp::AST const  & vars(args[0]);

      int  numVars(vars.nbArg());

      // restriction (b)
      if ( 1 >= numVars )
         throw std::runtime_error("notAllEqual restriction (b) error");

      // we impose the constraint imposing that at least one desequation
      // v[0] != v[i], for i in [1..numVars-1], is true
      // ( true == (v[0] != v[1]) || ... || (v[0] != v[numVars-1]) )
      gec::IntVar iV0 = constr::getIntVarFromDVar(gVars, gSpace, vars[0]);

      gec::BoolVar  bVal(gec::expr(
            gSpace,
            iV0 != constr::getIntVarFromDVar(gVars, gSpace, vars[1])));

      for ( int i = 2 ; i < numVars ; ++i )
         bVal = gec::expr(
            gSpace,
            bVal
            || (iV0
                != constr::getIntVarFromDVar(gVars, gSpace, vars[i])));

      gec::rel(gSpace, bVal, gec::IRT_EQ, 1);
    }


   // ***********************************************************************
   // --- XCSP 2.1 [not in Global Constraint Catalogue]:
   // - Arguments:
   //    LIST  𝚌𝚘𝚕𝚕𝚎𝚌𝚝𝚒𝚘𝚗(coef−𝚒𝚗𝚝, var−𝚍𝚟𝚊𝚛)
   //    OP    (<eq/> | <ne/> | <ge/> | <gt/> | <le/> | <lt/>)
   //    VAL   𝚒𝚗𝚝
   // - Definition:
   //    "weightedSum" := "(sum(i=1..r, LIST[i].coef * LIST[i].var) OP VAL,
   //                      where 'r' denotes the arity of the constraint"
   // - Restrictions:
   //    (a) required(LIST, [coef, var])
   // ***********************************************************************
   void globWeightedSum::manage (cxp::ASTList const      & args,
                                 gec::IntVarArray const  & gVars,
                                 gec::Space              & gSpace) const
    {
      if ( 3 != args.nbArg() )
         throw std::runtime_error("weightedSum args cardinality error");

      cxp::AST const  & leftElems(args[0]);
      cxp::AST const  & oper(args[1]);
      cxp::AST const  & rightValue(args[2]);

      if ( false == leftElems.isList() ) 
         throw std::runtime_error("weightedSum arg 1 error");

      if ( false == oper.isSymbol() ) 
         throw std::runtime_error("weightedSum arg 2 error");

      gec::IntArgs     iA;
      gec::IntVarArgs  iVA;

      for ( int i=0 ; i < leftElems.nbArg() ; ++i )
       {
         // restriction (a) satisfied or exception
         iA  << leftElems[i]["coef"].getInteger();
         iVA << this->getIntVarFromDVar(gVars, gSpace,
                                        leftElems[i]["var"]);
       }

      gec::IntVar  vRV(this->getIntVarFromDVar(gVars, gSpace,
                                               rightValue));

      if ( vRV.assigned() )
         gec::linear(gSpace, iA, iVA, this->getRelOp(oper.getType()), 
                     vRV.val());
      else
         gec::linear(gSpace, iA, iVA, this->getRelOp(oper.getType()), vRV);
    }
 }

