
#ifndef  __xcsp4gecode_constraints_h__
#define  __xcsp4gecode_constraints_h__  (1)


#include <xcsp-parser/XMLParser_libxml2.hh>

#include <gecode/int.hh>
#include <gecode/minimodel.hh>

namespace gec = Gecode;
namespace cxp = CSPXMLParser;

namespace x4g
 {
   class  constr
    {
      private:

         static std::map<NodeType, gec::BoolOpType> const  mOpBT;
         static std::map<NodeType, gec::IntRelType> const  mOpRT;
         static std::map<NodeType, gec::IntRelType> const  mOpRTRev;

      protected:

         static std::map<NodeType, gec::BoolOpType>  initMOpBT ();
         static std::map<NodeType, gec::IntRelType>  initMOpRT ();
         static std::map<NodeType, gec::IntRelType>  initMOpRTRev ();

         static gec::BoolOpType  getBoolOp (NodeType  oper)
          {
            std::map<NodeType, gec::BoolOpType>::const_iterator
               f(constr::mOpBT.find(oper));

            if ( constr::mOpBT.end() == f )
               throw std::runtime_error("bool/bool operator failure");

            return(f->second);
          }

         static gec::IntRelType  getRelOp (NodeType  oper)
          {
            std::map<NodeType, gec::IntRelType>::const_iterator
               f(constr::mOpRT.find(oper));

            if ( constr::mOpRT.end() == f )
               throw std::runtime_error("int/bool operator failure");

            return(f->second);
          }

         static gec::IntRelType  getRevRelOp (NodeType  oper)
          {
            std::map<NodeType, gec::IntRelType>::const_iterator
               f(constr::mOpRTRev.find(oper));

            if ( constr::mOpRTRev.end() == f )
               throw std::runtime_error("int/bool operator failure");

            return(f->second);
          }

         static gec::BoolVar  getBoolVarFromBool (gec::Space      & gSpace,
                                                  cxp::AST const  & bVar)
          {
            bool val(bVar.getBoolean());

            return( gec::BoolVar(gSpace, val, val) );
          }

         static gec::IntVar  getIntVarFromInt (gec::Space      & gSpace,
                                               cxp::AST const  & dVar)
          {
            int val(dVar.getInteger());

            return( gec::IntVar(gSpace, val, val) );
          }

         static gec::IntVar  getIntVarFromDVar
            (gec::IntVarArray const  & gVars,
             gec::Space              & gSpace,
             cxp::AST const          & dVar)
             {
               gec::IntVar  ret;

               if ( dVar.isVar() )
                  ret = gVars[dVar.getVarId()];
               else if ( dVar.isInteger() )
                {
                  int val(dVar.getInteger());

                  ret = gec::IntVar(gSpace, val, val);
                }
               else
                  throw std::runtime_error("getIntVarFromDVar type error");

               return(ret);
             }

      public:

         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const = 0;
    };


   class rel : public constr
    {
      private:

         RelType           relType;
         int               arity;
         int               numTuples;
         bool              finalized;
         gec::TupleSet     tSet;
         std::vector<int>  cV;

         // private default costructor to prevent the use
         rel ()
          { }

         static gec::IntVar getIntVarFromVar
            (gec::IntVarArray const  & gVars,
             cxp::AST const          & varId)
             {
               if ( 1 != varId.isVar() )
                  throw std::runtime_error("rel varId failure");

               return(gVars[varId.getVarId()]);
             }

         static gec::BoolVar getPostAnd (cxp::ASTList const      & args,
                                         gec::IntVarArray const  & gVars,
                                         gec::Space              & gSpace,
                                         int const *               tuple);

         void manageConf (cxp::ASTList const      & args,
                          gec::IntVarArray const  & gVars,
                          gec::Space              & gSpace) const;

         void manageSupp (cxp::ASTList const      & args,
                          gec::IntVarArray const  & gVars,
                          gec::Space              & gSpace) const
          {
            gec::IntVarArgs iA;

            for ( int i = 0 ; i < this->arity ; ++i )
               iA << this->getIntVarFromVar(gVars, args[i]);

            gec::extensional(gSpace, iA, this->tSet);
          }

         // Warning: don't change the start of the throw message (must
         // begin with "unsupported " for a CSC-09 compliant exit
         // message)
         void manageSoft (cxp::ASTList const      & args,
                          gec::IntVarArray const  & gVars,
                          gec::Space              & gSpace) const
          { throw std::runtime_error("unsupported soft relation"); }

      public:

         rel (RelType rt0, int a0, int nt0)
            : relType(rt0), arity(a0), numTuples(nt0), finalized(false)
             {
               if ( 0 >= this->arity )
                  throw std::runtime_error("relation positivity error");

               if ( REL_SOFT == this->relType )
                  this->cV.reserve(this->numTuples);
             }

         void add (int a0, int * t0, int c0 = 1);

         void finalize ();

         int getArity () const
          { return(this->arity); }

         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };


   class pred : public constr
    {
      private:
         static std::map<NodeType, int>  const  mOpA;

         bool                        noOper;
         bool                        valNoOper;
         int                         numArgs;     
         int                         operArgs;
         NodeType                    oper;
         cxp::AST *                  astTree;
         cxp::ASTAbstractFunction *  astFunc;


         static std::map<NodeType, int>  initMOpA ();

         static int  getNArgs (NodeType  oper)
          {
            std::map<NodeType, int>::const_iterator
               f(pred::mOpA.find(oper));

            if ( pred::mOpA.end() == f )
               throw std::runtime_error("unexpected operator id");

            return(f->second);
          }

         static bool  isBooleanOp (NodeType  oper)
          {
            return(   (F_NOT == oper) || (F_EQ  == oper) || (F_NE == oper)
                   || (F_GE  == oper) || (F_GT  == oper) || (F_LE == oper)
                   || (F_LT  == oper) || (F_AND == oper) || (F_OR == oper)
                   || (F_XOR == oper) || (F_IFF == oper) ); 
          }

         static bool  isIntegerOp (NodeType  oper)
          {
            return(   (F_NEG == oper) || (F_ABS == oper) || (F_ADD == oper)
                   || (F_SUB == oper) || (F_MUL == oper) || (F_DIV == oper)
                   || (F_MOD == oper) || (F_POW == oper) || (F_MIN == oper)
                   || (F_MAX == oper) || (F_IF  == oper) );
          }

         static bool  isBooleanOp2BoolArgs (NodeType  oper)
          {
            return(   (F_AND == oper) || (F_OR == oper) || (F_XOR == oper)
                   || (F_IFF == oper) ); 
          }

         static bool  isBooleanOp2IntArgs (NodeType  oper)
          {
            return(   (F_EQ == oper) || (F_NE == oper) || (F_GE == oper)
                   || (F_GT == oper) || (F_LE == oper) || (F_LT == oper) ); 
          }

         static gec::IntVar getIntVarFromExpr
            (cxp::ASTList const      & args,
             gec::IntVarArray const  & gVars, 
             gec::Space              & gSpace,
             cxp::AST const          & expr);

         static gec::BoolVar getBoolVarFromExpr
            (cxp::ASTList const      & args,
             gec::IntVarArray const  & gVars, 
             gec::Space              & gSpace,
             cxp::AST const          & expr);

         static gec::IntVar getPowVarFromExpr
            (cxp::ASTList const      & args,
             gec::IntVarArray const  & gVars, 
             gec::Space              & gSpace,
             cxp::AST const          & expr,
             int                       valExp);

         void checkAndSetAstFunc ();
         void manageInt2 (cxp::ASTList const      & args,
                          gec::IntVarArray const  & gVars,
                          gec::Space              & gSpace) const;

         void manageNot (cxp::ASTList const      & args,
                         gec::IntVarArray const  & gVars,
                         gec::Space              & gSpace) const
          {
            gec::rel(
               gSpace,
               this->getBoolVarFromExpr(args, gVars, gSpace,
                                        this->astFunc->getArg(0)),
               gec::IRT_NQ,
               1);
          }

         void manageBool2 (cxp::ASTList const      & args,
                           gec::IntVarArray const  & gVars,
                           gec::Space              & gSpace) const
          {
            gec::rel(
               gSpace,
               this->getBoolVarFromExpr(args, gVars, gSpace,
                                        this->astFunc->getArg(0)),
               this->getBoolOp(this->oper),
               this->getBoolVarFromExpr(args, gVars, gSpace,
                                        this->astFunc->getArg(1)),
               1);
          }

      public:
         pred ()
            : numArgs(0), astTree(NULL)
             { }

         pred (int  nA0, cxp::AST *  aT0)
            : noOper(false), valNoOper(false), numArgs(nA0), oper(F_EQ),
            astTree(aT0), astFunc(NULL)
             { this->checkAndSetAstFunc(); }

         ~pred ()
          { if ( this->astTree ) delete this->astTree; }

         void setNumArgs (int  nA0)
          { this->numArgs = nA0; }

         void setAstTree (cxp::AST *  aT0)
          { this->astTree = aT0; this->checkAndSetAstFunc(); }

         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };


   class globAllDifferent : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globAmong : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globAtBase : public constr
    {
      protected:
         static void verifAt (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace,
                              gec::IntRelType           gRel);
    };

   class globAtLeast : public globAtBase
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const
          { this->verifAt(args, gVars, gSpace, gec::IRT_GQ); }
    };

   class globAtMost : public globAtBase
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const
          { this->verifAt(args, gVars, gSpace, gec::IRT_LQ); }
    };

   class globCumulative : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globDiffn : public constr
    {
      private:
         typedef  enum { oOri, oSiz, oEnd }          orthCoords;
         typedef  std::map<orthCoords, gec::IntVar>  orthTriple;
         typedef  std::vector<orthTriple>            orth;

         static void loadTriple (gec::IntVarArray const  & gVars,
                                 gec::Space              & gSpace,
                                 globDiffn::orthTriple   & gdfT,
                                 cxp::AST const          & astT);

         static void loadOrthotope (gec::IntVarArray const  & gVars,
                                    gec::Space              & gSpace,
                                    globDiffn::orth         & gdfO,
                                    cxp::AST const          & astO,
                                    int                       numTpl);

         static void setDiffn4two (gec::IntVarArray const  & gVars,
                                   gec::Space              & gSpace,
                                   globDiffn::orth         & o1,
                                   globDiffn::orth         & o2);

      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globDisjunctive : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globElement : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globGlobalCardinality : public constr
    {
      private:
         static gec::IntVar getGCExpr (gec::Space             & gSpace,
                                       gec::IntVarArgs const  & iVA,
                                       int                      val);

      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globLex : public constr
    {
      protected:
         void setLex (cxp::ASTList const      & args,
                      gec::IntVarArray const  & gVars,
                      gec::Space              & gSpace,
                      gec::IntRelType           gOp) const;
    };

   class globLexLess : public globLex
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const
          { this->setLex(args, gVars, gSpace, gec::IRT_LE); }
    };

   class globLexLessEq : public globLex
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const
          { this->setLex(args, gVars, gSpace, gec::IRT_LQ); }
    };

   class globNotAllEqual : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };

   class globWeightedSum : public constr
    {
      public:
         virtual void manage (cxp::ASTList const      & args,
                              gec::IntVarArray const  & gVars,
                              gec::Space              & gSpace) const;
    };
 }

#endif  // __xcsp4gecode_constraints_h__

