
#include <x4g/parser.h>


namespace x4g
 {
   std::map<std::string, x4g::constr const *> XCSPParser::initGC ()
    {
      std::map<std::string, x4g::constr const *>  ret;

      ret["global:alldifferent"]       = new x4g::globAllDifferent();
      ret["global:among"]              = new x4g::globAmong();
      ret["global:atleast"]            = new x4g::globAtLeast();
      ret["global:atmost"]             = new x4g::globAtMost();
      ret["global:cumulative"]         = new x4g::globCumulative();
      ret["global:diffn"]              = new x4g::globDiffn();
      ret["global:disjunctive"]        = new x4g::globDisjunctive();
      ret["global:element"]            = new x4g::globElement();
      ret["global:global_cardinality"] = new x4g::globGlobalCardinality();
      ret["global:lexless"]            = new x4g::globLexLess();
      ret["global:lexlesseq"]          = new x4g::globLexLessEq();
      ret["global:not_all_equal"]      = new x4g::globNotAllEqual();
      ret["global:weightedsum"]        = new x4g::globWeightedSum();

      return(ret);
    }

   XCSPParser::~XCSPParser ()
    {
      std::vector<x4g::rel *>::iterator  rit;

      for ( rit = this->rels.begin() ; this->rels.end() != rit ; ++rit )
         delete *rit;

      std::vector<x4g::pred *>::iterator  pit;

      for ( pit = this->preds.begin() ; this->preds.end() != pit ; ++pit )
         delete *pit;
    }

   void XCSPParser::beginDomain (std::string const &  name,
                                 int                  idDomain,
                                 int                  nbValue) 
    {
      if ( this->counter != idDomain )
         throw std::runtime_error("id domain error");

      if ( this->counter >= int(this->domainSets.size()) )
         throw std::runtime_error("too much domains");

      if ( nbValue <= 0 )
         throw std::runtime_error("domain with illegal num values");

      this->domainNames[this->counter] = name;
      this->num                        = nbValue;
      this->intArgs                    = gec::IntArgs();
    }

   void XCSPParser::addVariable (std::string const &  name,
                                 int                  idVar,
                                 std::string const &  domain,
                                 int                  idDomain) 
    {
      if ( this->counter != idVar )
         throw std::runtime_error("id var error");

      if ( this->counter >= int(this->varNames.size()) )
         throw std::runtime_error("too much vars");

      if ( this->domainNames[idDomain] != domain )
         throw std::runtime_error("domain id/name mismatch");

      this->varNames[this->counter]  = name;
      (*this->pGVars)[this->counter] = 
         gec::IntVar(*(this->pGSpace), this->domainSets[idDomain]);

      ++ this->counter;
    }

   void XCSPParser::beginConstraint (std::string const       & name,
                                     int                       idConstr,
                                     int                       arity, 
                                     std::string const       & reference, 
                                     cxp::CSPDefinitionType    type,
                                     int                       id,
                                     cxp::ASTList const      & scope)
    {
      if ( this->counter != idConstr )
         throw std::runtime_error("id constraint error");

      if ( this->counter >= this->numConstrs )
         throw std::runtime_error("too much constraints");

      switch ( type )
       {
         case cxp::RelationType:
            if ( (id < 0) || (id >= int(this->rels.size())) )
               throw std::runtime_error("relation id failure");

            this->activeConstr = this->rels[id];

            if ( this->rels[id]->getArity() != arity )
               throw std::runtime_error("relation arity error");
            break;

         case cxp::PredicateType:
            if ( (id < 0) || (id >= int(this->preds.size())) )
               throw std::runtime_error("predicate id failure");

            this->activeConstr = this->preds[id];
            break;

         case cxp::GlobalConstraintType:
             {
               std::map<std::string, x4g::constr const *>::const_iterator
                  ci(XCSPParser::globalConstrs.find(reference));

               // Warning: don't change the start of the throw message (must
               // begin with "unsupported " for a CSC-09 compliant exit
               // message)
               if ( XCSPParser::globalConstrs.end() == ci )
                  throw std::runtime_error("unsupported global constraint");

               this->activeConstr = ci->second;
             }
            break;

         default:
            throw std::runtime_error("unexpected constraint type");
            break;
       }
    }

   std::map<std::string, x4g::constr const *> const
      XCSPParser::globalConstrs(XCSPParser::initGC());

 } // end namespace x4g

