
#ifndef  __xcsp4gecode_parser_h__
#define  __xcsp4gecode_parser_h__  (1)


#include <xcsp-parser/XMLParser_libxml2.hh>

#include <x4g/constraints.h>

namespace gec = Gecode;
namespace cxp = CSPXMLParser;

namespace x4g
 {
   class XCSPParser : public cxp::CSPParserCallback
    {
      private:

         static std::map<std::string, x4g::constr const *> const
            globalConstrs;

         int                       numConstrs;
         int                       numRels;
         int                       numPreds;
         int                       num;
         int                       counter;
         x4g::constr const *       activeConstr;
         std::vector<std::string>  domainNames;
         std::vector<std::string>  varNames;
         std::vector<gec::IntSet>  domainSets;
         std::vector<x4g::rel *>   rels;
         std::vector<x4g::pred *>  preds;
         gec::IntArgs              intArgs;
         gec::Space *              pGSpace;
         gec::IntVarArray *        pGVars;


         static std::map<std::string, x4g::constr const *> initGC ();

         XCSPParser ()
          { }


      public:

         XCSPParser (gec::Space * pGSpace0, gec::IntVarArray * pGVars0)
            : numConstrs(0), numRels(0), numPreds(0), num(0), counter(0),
            activeConstr(NULL), pGSpace(pGSpace0), pGVars(pGVars0)
             { }

         ~XCSPParser ();

         std::vector<std::string> const & getVarNames () const
          { return(this->varNames); }

         virtual void beginInstance (const std::string &  name)
          { }

         virtual void beginDomainsSection (int  nbDomains) 
          {  
            this->counter = 0;

            this->domainNames.resize(nbDomains);
            this->domainSets.resize(nbDomains);
          }

         virtual void beginDomain (std::string const &  name,
                                   int                  idDomain,
                                   int                  nbValue);

         virtual void addDomainValue (int v) 
          { this->intArgs << v; }

         virtual void addDomainValue (int  first,
                                      int  last) 
          {
            if ( first > last )
               throw std::runtime_error("illegal domain range");

            this->intArgs << gec::IntArgs().create(last-first+1, first);
          }

         virtual void endDomain () 
          {
            this->domainSets[this->counter] = gec::IntSet(this->intArgs);

            if ( this->num != (int)this->domainSets[this->counter].size() )
               throw std::runtime_error("domain cartdinality error");

            ++ this->counter;
          }

         virtual void endDomainsSection () 
          {
            if ( this->counter != int(this->domainSets.size()) )
               throw std::runtime_error("domains cardinality error");
          }

         virtual void beginVariablesSection (int  nbVariables) 
          {
            this->counter   = 0;
            *(this->pGVars) = gec::IntVarArray(*(this->pGSpace),
                                                nbVariables);

            this->varNames.resize(nbVariables);
          }

         virtual void addVariable (std::string const &  name,
                                   int                  idVar,
                                   std::string const &  domain,
                                   int                  idDomain);

         virtual void endVariablesSection () 
          {
            if ( this->counter != int(this->varNames.size()) )
               throw std::runtime_error("variables cardinality error");
          }

         virtual void beginRelationsSection (int  nbRelations) 
          {
            this->numRels = nbRelations;
            this->counter = 0;

            this->rels.reserve(this->numRels);
          }

         virtual void beginRelation (std::string const  & name,
                                     int                  idRel,
                                     int                  arity,
                                     int                  nbTuples,
                                     RelType              relType) 
          {
            if ( this->counter != idRel )
               throw std::runtime_error("id relation error");

            if ( this->counter >= this->numRels )
               throw std::runtime_error("too much relations");

            this->rels.push_back(new x4g::rel(relType, arity, nbTuples));
          }

         virtual void addRelationTuple (int  arity,
                                        int  tuple[]) 
          { this->rels.back()->add(arity, tuple); }

         virtual void addRelationTuple (int  arity,
                                        int  tuple[],
                                        int  cost) 
          { this->rels.back()->add(arity, tuple, cost); }

         virtual void endRelation () 
          {
            this->rels.back()->finalize();

            ++ this->counter;
          }

         virtual void endRelationsSection () 
          {
            if ( this->counter != this->numRels )
               throw std::runtime_error("relations cardinality error");
          }

         virtual void beginPredicatesSection (int  nbPredicates) 
          {
            this->numPreds = nbPredicates;
            this->counter  = 0;

            this->preds.reserve(this->numPreds);
          }

         virtual void beginPredicate (std::string const  & name,
                                      int                  idPred) 
          {
            if ( this->counter != idPred )
               throw std::runtime_error("id predicate error");

            if ( this->counter >= this->numPreds )
               throw std::runtime_error("too much predicates");

            this->num = 0;
          }

         virtual void addFormalParameter (int                  pos,
                                          std::string const  & name, 
                                          std::string const  & type) 
          {
            if ( pos != this->num )
               throw std::runtime_error("id formal parameter error");

            if ( type != "int" )
               throw std::runtime_error("type formal parameter error");

            ++ this->num;
          }

         virtual void predicateExpression (cxp::AST *  tree) 
          { this->preds.push_back(new x4g::pred(this->num, tree)); }

         virtual void endPredicate () 
          { ++ this->counter; }

         virtual void endPredicatesSection () 
          {
            if ( this->counter != this->numPreds )
               throw std::runtime_error("predicates cardinality error");
          }

         virtual void beginConstraintsSection (int  nbConstraints) 
          {
            this->numConstrs = nbConstraints;
            this->counter    = 0;
          }

         virtual void beginConstraint (std::string const       & name,
                                       int                       idConstr,
                                       int                       arity, 
                                       std::string const       & reference, 
                                       cxp::CSPDefinitionType    type,
                                       int                       id,
                                       cxp::ASTList const      & scope);

         virtual void constraintParameters (cxp::ASTList const  & args)
          {
            this->activeConstr->manage(args, *(this->pGVars),
                                       *(this->pGSpace));
          } 

         virtual void endConstraint () 
          { ++ this->counter; }

         virtual void endConstraintsSection () 
          {
            if ( this->counter != this->numConstrs )
               throw std::runtime_error("constraints cardinality error");
          }

         virtual void endInstance () 
          { }
    };

 }  // end namespace x4g

#endif // ndef __xcsp4gecode_parser_h__

